/***************************************************************************
 *   Copyright (C) 2004-2005 by Andrew Krause                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "notebookmanager.h"

// Notebook manager constructor that creates the notebook and sets its
// properties. It also sets the number of tabs to zero
NotebookManager::NotebookManager()
{
  // Create the notebook & set its settings
  notebook = gtk_notebook_new();
  gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), true);
  gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), GTK_POS_TOP);
  gtk_notebook_set_scrollable(GTK_NOTEBOOK(notebook), true);
  gtk_widget_show(notebook);

  g_signal_connect (G_OBJECT(notebook), "change_current_page",
                    G_CALLBACK(notebook_event), (gpointer)this);
  tab_count = 0;
}

// Catch all events of the notebook so that the focus of the textview can
// be restored when a notebook tab is clicked
void notebook_event (GtkNotebook *widget,
                     gint page,
                     gpointer data)
{
  NotebookManager *ptr = (NotebookManager*) data;
  gint current_page = gtk_notebook_get_current_page(GTK_NOTEBOOK(widget));
  gtk_widget_grab_focus(ptr->tabs[current_page].textview);
}

// This function allows the main openldev class direct access to
// the notebook widget
GtkWidget* NotebookManager::get_notebook()
{ return notebook;
}

// Add a tab to the notebook. It receives the filename that is to be opened
// and the environment settings object. A new tab object is created if the
// file is not already open. The signals are then connected to the widgets and
// the fonts are set to the line numbering & TextView
bool NotebookManager::add_tab (string filename, 
                               EnvironmentSettings *env)
{
  NotebookTab tab;
  tab.fn = filename;
  
  // Parse the filename to remove directory structure
  while (filename.find('/',0) != string::npos)
    filename.erase(0,filename.find('/',0)+1);

  // Check to make sure file is not already open
  for (int i=0; i<tab_count; i++)
  {    
    if (tab.fn == tabs[i].fn)
    {
      gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), i);
      return false;
    }
  }

  // Create temporary tab & initialize the objects
	tab.initialize_objects(filename);
	tab.tab_id = tab_count;

  // Add the tab to the notebook and notebook_tab's deque
	tabs.push_back(tab);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
                           tabs[tab_count].swin,
                           tabs[tab_count].hbox);
  
  // Set all of the signal commands
  g_signal_connect (G_OBJECT(tabs[tab_count].button), "clicked",
                    G_CALLBACK(delete_tab), (gpointer)this);
  g_signal_connect (G_OBJECT(tabs[tab_count].textview), "key-press-event",
                    G_CALLBACK(on_keypress), (gpointer)this);
   
  // Set the font and tab size of the textview  
  string font_description = env->get_font_face() + " " + env->get_font_size();
  PangoFontDescription *fd = pango_font_description_from_string(font_description.c_str());
  gtk_text_view_set_tabs(GTK_TEXT_VIEW(tabs[tab_count].textview), 
                         get_tab_array(fd, env, tabs[tab_count].textview));
  
  // Do final initialization
  gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), tab_count);
	tab_count++;
  return true;
}

// Revert to the saved version of the file. The file must be set as unmodified,
// then the file is reread & restored to its original version.
void NotebookManager::revert_to_saved()
{
  mark_current_tab_modified(false);
  
  int current = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
  string text = read_file(tabs[current].fn);
  
  GtkTextBuffer *buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(tabs[current].textview));
  gtk_text_buffer_set_text(buffer, text.c_str(), -1);
  gtk_text_view_set_buffer(GTK_TEXT_VIEW(tabs[current].textview), GTK_TEXT_BUFFER(buffer));
}

// Delete the currently selected tab. First the user must confirm a closing of
// the file if it is modified. Then find the close button that was clicked,
// remove the tab, and clean up after the tab. It receives the full filename
// of the file that is going to be closed.
void delete_tab (GtkWidget *widget,
                 gpointer data)
{
  NotebookManager *ptr = (NotebookManager*) data;
  
  for (int i=0; i < ptr->tab_count; i++)
    if (ptr->tabs[i].button == widget)
    {
      string label_text = gtk_label_get_text(GTK_LABEL(ptr->tabs[i].label));
     
      if (label_text.find('*', 0) != string::npos)
        if (!question("The file you are closing is not saved.\nAre you sure you want to close?"))
          return;

      vector<NotebookTab>::iterator itr = ptr->tabs.begin();
      gtk_notebook_remove_page(GTK_NOTEBOOK(ptr->notebook), i);

      for (int count=0; count < i; count++)
        itr++;

      ptr->tab_count--;
      ptr->tabs.erase(itr);

	    if (ptr->tab_count != 0)
        gtk_notebook_set_current_page(GTK_NOTEBOOK(ptr->notebook), i-1);
      break;
    }    
}

// Return whether there is currently a file open. This will prevent a
// segmentation fault when compiling
bool NotebookManager::get_file_open()
{
  if (gtk_notebook_get_n_pages(GTK_NOTEBOOK(notebook)) > 0)
    return true;
  
  return false;
}

// Save all files in the tabs. Loop through all of the files and set them as
// unmodified. Then write the text from the buffer to the tab's full filename.
void NotebookManager::save_all_files()
{
  GtkTextIter start, end;
  
  for (int i=0; i<gtk_notebook_get_n_pages(GTK_NOTEBOOK(notebook)); i++)
  {    
    // Set the label as unmodified
    string text = gtk_label_get_text(GTK_LABEL(tabs[i].label));
    if (text.find('*', 0) != string::npos)
    {
      text.erase(text.find('*', 0), 1);
      gtk_label_set_text(GTK_LABEL(tabs[i].label), text.c_str());
    
      // Save the file if it is unmodified
      GtkTextBuffer *buffer = GTK_TEXT_BUFFER(gtk_text_view_get_buffer(GTK_TEXT_VIEW(tabs[i].textview)));
      write_file(tabs[i].fn, gtk_text_buffer_get_text(buffer, &start, &end, true));
    }
  }
}

// Destroy everything in the notebook manager. This will remove all of the
// pages until the there are no more left.
void NotebookManager::erase_all_tabs()
{
  // Remove all of the tabs and tab information
  tabs.erase(tabs.begin(), tabs.end());
  while (gtk_notebook_get_n_pages(GTK_NOTEBOOK(notebook)) != 0)
    gtk_notebook_remove_page(GTK_NOTEBOOK(notebook), 0);
  
  tab_count = 0;
}

// Return the current tab so that it can be edited by the main openldev class
int NotebookManager::get_current_tab()
{ return gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
}

bool on_keypress (GtkWidget *widget,
                  GdkEventKey *event,
                  gpointer data)
{
  NotebookManager *ptr = (NotebookManager*) data;
  if (event->type == GDK_KEY_PRESS) 
    ptr->mark_current_tab_modified(true);
    
  return false;
}

// Mark the current file as modified or unmodified. It receives whether it should
// be marked as modified or unmodified. Modified is shown by adding an asterick
// to the end of the short filename in the tab header
void NotebookManager::mark_current_tab_modified (bool modified)
{
  int current = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
  string text = gtk_label_get_text(GTK_LABEL(tabs[current].label));
  
  // If the file was modified
  if (modified)
  {
    if (text.find('*', 0) == string::npos)
      text += "*";
  }
  // If it was not modified
  else 
  {
    if (text.find('*', 0) != string::npos)
      text.erase(text.find('*', 0), 1);
  }
  
  gtk_label_set_text(GTK_LABEL(tabs[current].label), text.c_str());
}


// Modify the font of each textview & line numbering. This receives a font
// description and is used when the user changes their default font.
void NotebookManager::modify_fonts (PangoFontDescription *fd)
{
  // Foreach tab, modify the fonts
  for (int i=0; i<tab_count; i++)
    gtk_widget_modify_font(tabs[i].textview, fd);
}

// Scroll to the current cursor position. This is used for keypresses, find
// commands and goto line commands. It is needed for keypresses because the line
// numbering messes up the default properties of the TextView such as scrolling
// to the cursor when it reaches the end of the screen.
void NotebookManager::scroll_to_mark()
{
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(get_textbuffer());
  GtkTextIter iter;
  gtk_text_buffer_get_iter_at_mark(buffer, &iter, gtk_text_buffer_get_insert(buffer));
  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(get_textview()), &iter, 0.0, false, 0.0, 0.0);
}

// Get the filename of the current tab. This returns the full filename.
string NotebookManager::get_current_filename()
{ return tabs[gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook))].fn;
}

// Get the short filename of the current tab. The asterick must be removed from
// the end of modified file names.
string NotebookManager::get_current_short_filename()
{
  int current = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
  string fn = gtk_label_get_text(GTK_LABEL(tabs[current].label));
  if (fn.find("*", 0) != string::npos)
    fn.erase(fn.find("*", 0), 1);
  
  return fn;
}

// Return the current textview so the main openldev class can directly access
// it. If there are no files open, then return a brand-spanking new textview
// that can just be trashed.
GtkWidget* NotebookManager::get_textview()
{
  int current = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));

  // Return the current textview
  if (tab_count != 0)
    return tabs[current].textview;
  // Return an empty textview if there are none
  else
  {
    GtkWidget *textview = gtk_source_view_new();
    return textview;
  }
}

// Return the current buffer so the main openldev class can directly access it. 
GtkSourceBuffer* NotebookManager::get_textbuffer()
{ return GTK_SOURCE_BUFFER(gtk_text_view_get_buffer(GTK_TEXT_VIEW(get_textview())));
}

// Return the offset of the cursor iterator
int NotebookManager::get_insert_offset()
{ 
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(get_textbuffer());
  GtkTextIter iter;
  
  gtk_text_buffer_get_iter_at_mark(buffer, &iter,
                                   gtk_text_buffer_get_insert(buffer));
  return gtk_text_iter_get_offset(&iter);
}

// Return the offset of the selection bound iterator
int NotebookManager::get_selection_offset()
{ 
  GtkTextBuffer *buffer = GTK_TEXT_BUFFER(get_textbuffer());
  GtkTextIter iter;
  
  gtk_text_buffer_get_iter_at_mark(buffer, &iter,
                                   gtk_text_buffer_get_selection_bound(buffer));
  return gtk_text_iter_get_offset(&iter);
}

string get_mime_type(string fn)
{
  gnome_vfs_init();
  return gnome_vfs_get_mime_type(fn.c_str());
}

void NotebookManager::set_properties_one_tab (int tab, 
                                              EnvironmentSettings* env)
{
  GtkSourceBuffer *buffer = GTK_SOURCE_BUFFER(gtk_text_view_get_buffer(GTK_TEXT_VIEW(tabs[tab].textview)));
  
  gtk_source_view_set_show_line_numbers(GTK_SOURCE_VIEW(tabs[tab].textview),
                                        env->get_line_numbers());
  gtk_source_view_set_tabs_width(GTK_SOURCE_VIEW(tabs[tab].textview),
                                 env->get_tabsize());
  gtk_source_view_set_auto_indent(GTK_SOURCE_VIEW(tabs[tab].textview),
                                  env->get_auto_indent());
  gtk_source_view_set_show_margin(GTK_SOURCE_VIEW(tabs[tab].textview),
                                  env->get_rightmargin());
  gtk_source_view_set_insert_spaces_instead_of_tabs(GTK_SOURCE_VIEW(tabs[tab].textview),
                                                    env->get_replace_tabs());
  gtk_source_buffer_set_max_undo_levels(buffer, env->get_undo());
  gtk_source_buffer_set_check_brackets(buffer, env->get_brackets());

  string mime_type = get_mime_type(tabs[tab].fn); 
  if (mime_type == "NONE")
    gtk_source_buffer_set_highlight(buffer, false);
  else
  {
    gtk_source_buffer_set_highlight(buffer, true);
    GtkSourceLanguagesManager *manager = gtk_source_languages_manager_new();
    GtkSourceLanguage* language = gtk_source_languages_manager_get_language_from_mime_type(manager, mime_type.c_str());

    if (language)
      gtk_source_buffer_set_language(buffer, language);
  }
  
  if (env->get_wrap())
  {
    if (env->get_wrapsplit())
      gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(tabs[tab].textview), GTK_WRAP_CHAR);
    else
      gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(tabs[tab].textview), GTK_WRAP_WORD);
  }
  else
      gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(tabs[tab].textview), GTK_WRAP_NONE);
}

void NotebookManager::set_properties (bool all_files, 
                                      EnvironmentSettings* env)
{
  if (all_files)
    for (int i=0; i<tab_count; i++)
      set_properties_one_tab(i, env);
  else
    set_properties_one_tab(gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook)), env);
}

void NotebookManager::set_filename (string file)
{
  while (file.find('/',0) != string::npos)
    file.erase(0, file.find('/',0) + 1);
  
  int current = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
  gtk_label_set_text(GTK_LABEL(tabs[current].label), file.c_str());
}
